use crate::{
    errors::NoUserInfo,
    handlers::{markups::deletion_markup, Handler, MainDialogue, State},
    models::User,
};
use futures::{future, stream::FuturesUnordered, StreamExt};
use itertools::Itertools;
use sea_orm::prelude::*;
use std::sync::Arc;
use teloxide::{adaptors::Throttle, prelude::*};
use tokio::task::spawn_blocking;

/// Gets the master password, encryptes and adds the accounts to the DB
async fn get_master_pass(
    bot: Throttle<Bot>,
    msg: Message,
    db: DatabaseConnection,
    dialogue: MainDialogue,
    master_pass: String,
    user: User,
) -> crate::Result<()> {
    let user_id = msg.from().ok_or(NoUserInfo)?.id.0;
    let accounts = user.accounts;

    let failed: Vec<String> = {
        let master_pass: Arc<str> = master_pass.into();
        let db = &db;

        let futures: FuturesUnordered<_> = accounts
            .into_iter()
            .map(|account| {
                let master_pass = Arc::clone(&master_pass);
                async move {
                    if !account.validate() {
                        return Err(account.name);
                    }
                    let name = account.name.clone();
                    match spawn_blocking(move || account.into_account(user_id, &master_pass)).await
                    {
                        Ok(Ok(account)) => match account.insert(db).await {
                            Ok(_) => Ok(()),
                            Err(_) => Err(name),
                        },
                        _ => Err(name),
                    }
                }
            })
            .collect();

        futures
            .filter_map(|result| future::ready(result.err()))
            .collect()
            .await
    };
    let message = if failed.is_empty() {
        "Success".to_owned()
    } else {
        format!(
            "Failed to create the following accounts:\n{}",
            failed.into_iter().format("\n")
        )
    };
    bot.send_message(msg.chat.id, message)
        .reply_markup(deletion_markup())
        .await?;
    dialogue.exit().await?;
    Ok(())
}

/// Downloads and parses the json and asks for the master password
async fn get_document(
    bot: Throttle<Bot>,
    msg: Message,
    _: DatabaseConnection,
    dialogue: MainDialogue,
    user: User,
) -> crate::Result<()> {
    let previous = bot
        .send_message(msg.chat.id, "Send master password")
        .await?;
    dialogue
        .update(State::GetMasterPass(Handler::new(
            move |bot, msg, db, dialogue, master_pass| {
                get_master_pass(bot, msg, db, dialogue, master_pass, user)
            },
            previous,
        )))
        .await?;
    Ok(())
}

/// Handles /import command
pub async fn import(bot: Throttle<Bot>, msg: Message, dialogue: MainDialogue) -> crate::Result<()> {
    let previous = bot
        .send_message(
            msg.chat.id,
            "Send a json document with the same format as created by /export",
        )
        .await?;
    dialogue
        .update(State::GetUser(Handler::new(get_document, previous)))
        .await?;
    Ok(())
}
