use crate::{
    entity::prelude::Account,
    errors::{HandlerUsed, NoUserInfo},
    handlers::{
        markups::{account_markup, deletion_markup},
        utils::{delete_optional, validate_field},
        MainDialogue,
    },
};
use sea_orm::prelude::*;
use teloxide::{adaptors::Throttle, prelude::*};

/// Specifies the kind of checks to be run during the account name validation
#[derive(Clone, Copy)]
pub enum NameCheckKind {
    NewAccountName,
    MustExist,
}

/// Validates the account name
#[inline]
async fn check_name(
    bot: &Throttle<Bot>,
    msg: &Message,
    db: &DatabaseConnection,
    name: &str,
    check_kind: NameCheckKind,
    user_id: u64,
) -> crate::Result<Option<Message>> {
    match check_kind {
        NameCheckKind::MustExist => {
            if !Account::exists(user_id, name, db).await? {
                let msg = bot
                    .send_message(msg.chat.id, "Account doesn't exists. Try again")
                    .reply_markup(account_markup(user_id, db).await?)
                    .await?;
                return Ok(Some(msg));
            }
        }
        NameCheckKind::NewAccountName => {
            if Account::exists(user_id, name, db).await? {
                let msg = bot
                    .send_message(msg.chat.id, "Account already exists")
                    .await?;
                return Ok(Some(msg));
            }
            if !validate_field(name) {
                let msg = bot
                    .send_message(msg.chat.id, "Invalid account name. Try again")
                    .await?;
                return Ok(Some(msg));
            }
        }
    }
    Ok(None)
}

/// Function to handle GetAccountName state
pub async fn get_account_name(
    bot: Throttle<Bot>,
    msg: Message,
    db: DatabaseConnection,
    dialogue: MainDialogue,
    (next, check_kind): (super::PackagedHandler<String>, NameCheckKind),
) -> crate::Result<()> {
    let user_id = msg.from().ok_or(NoUserInfo)?.id.0;
    let mut handler = next.lock().await;
    delete_optional(&bot, handler.previous.as_ref()).await;

    let text = match msg.text() {
        Some(text) => text.trim_end(),
        None => {
            let mut send = bot.send_message(
                msg.chat.id,
                "Couldn't get the text of the message. Send the name again",
            );
            if let NameCheckKind::MustExist = check_kind {
                send = send.reply_markup(account_markup(user_id, &db).await?)
            }
            send.await?;
            return Ok(());
        }
    };

    if text == "/cancel" {
        dialogue.exit().await?;
        bot.send_message(msg.chat.id, "Successfully cancelled")
            .reply_markup(deletion_markup())
            .await?;
        return Ok(());
    }

    if handler.func.is_none() {
        let _ = dialogue.exit().await;
        return Err(HandlerUsed.into());
    }

    if let Some(failure_message) = check_name(&bot, &msg, &db, text, check_kind, user_id).await? {
        handler.previous = Some(failure_message);
        return Ok(());
    }

    let func = handler.func.take().unwrap();
    drop(handler);
    let text = text.to_owned();

    if let Err(err) = func(bot, msg, db, dialogue.clone(), text).await {
        let _ = dialogue.exit().await;
        return Err(err);
    }
    Ok(())
}
